{-# LANGUAGE OverloadedStrings #-}

module Gitea.Preview.Markdown.ChromaHighlight where

import Control.Exception (Exception(..), throw)
import Data.Either (fromRight)
import Data.List (singleton)
import Data.Maybe (listToMaybe)
import Data.Text (Text)
import System.Exit (ExitCode(..))
import System.IO (hClose)
import System.IO.Unsafe (unsafePerformIO)
import System.Process (CreateProcess(..), StdStream(..), createProcess, proc, waitForProcess)
import Text.Pandoc (Block(..), Format(..), Pandoc)
import Text.Pandoc.Walk (walk)

import qualified Data.Text as T
import qualified Data.Text.IO as T

chromaHighlight' :: Pandoc -> Pandoc
chromaHighlight' = walk chromaHighlight

chromaHighlight :: Block -> Block
chromaHighlight (CodeBlock attr@(_, classes, _) text) = Div attr $ singleton $ RawBlock (Format "html") $
    tryChromaHighlight (T.unpack <$> listToMaybe classes) text
chromaHighlight x = x

tryChromaHighlight :: Maybe String -> Text -> Text
tryChromaHighlight (Just language) text = fromRight (tryChromaHighlight Nothing text) $
    chromaHighlightText language text
tryChromaHighlight Nothing text = either throw id $ chromaHighlightText "plain" text

chromaHighlightText :: String -> Text -> Either ChromaError Text
chromaHighlightText language text = unsafePerformIO $ do
    (Just hIn, Just hOut, Just hErr, processHandle) <- createProcess $
        (proc "chroma" ["--html", "--html-only", "--lexer", language])
            {std_in = CreatePipe, std_out = CreatePipe, std_err = CreatePipe}
    T.hPutStrLn hIn text
    hClose hIn
    highlighted <- T.hGetContents hOut
    errors <- T.hGetContents hErr
    exitCode <- waitForProcess processHandle
    pure $ case exitCode of
      ExitSuccess -> Right highlighted
      ExitFailure err -> Left $ ChromaError err errors

data ChromaError = ChromaError Int Text deriving Show

instance Exception ChromaError where
    displayException (ChromaError exitCode errors) =
        "chroma exited with code '" <> show exitCode <> "' and error message:\n" <> T.unpack errors
